/*
append.c

This is a version of the DOS APPEND command, written
for the FreeDOS project (freedos.org).
append.c compiles with the Borland Turbo C compiler, version
2.01.

Written by: Jesse <slicer69@hotmail.com>
http://slicer69.tripod.com/code/

Please see the append.txt file for more information.
*/

#include <stdio.h>
#include <string.h>


/* Use for debugging test-runs only. */
/* #define DEBUG 1 */

/* Place to store the append path. */
#define APPEND_FILE_NAME "append.tmp"

/* The all important version number. */
#define VERSION "0.1"

/* Data to be stored in/from the append file. */
typedef struct
{
   int x_flag;            /* whether X flag was set */
   int environment_flag;  /* Whether data is stored in DOS environment */
   char *append_path;     /* The append path(s) */
} APPEND_STRUCT;


/* The function to find command line arguments. */
int Find_Arg(char *find_arg, char *value, int argc, char *argv[]);

/* A function to up-case an array of strings. */
void Upcase_Strings(char **str, int num_strings);


/* This function reads data from the append file into the
   append structure.
*/
int Read_Append_Data(FILE *append_file, APPEND_STRUCT *append_struct);

/* Write the append data to a file.
   Returns true or false */
int Write_Append_Data(FILE *append_file, APPEND_STRUCT *append_struct);

/*
This function tries to find an append data file.
*/
FILE *Find_Append_File();


int main(int argc, char *argv[])
{
  APPEND_STRUCT append_struct;
  FILE *append_file = NULL;             /* File storing append data */
  int status;
  int done_task = 0;      /* if append's task has been done */

  /* init append structure */
  append_struct.append_path = NULL;
  
  /* Upcase all command line arguments */
  Upcase_Strings(argv, argc);


  if ( Find_Arg("/?", NULL, argc, argv) )
  {
     /* Print help message. */
     printf("Append version %s\n", VERSION);
     printf("Append stores/retrieves search path(s) for data files.\n\n");
     printf("Usage: append [/x] [/e]\n");
     printf("Usage: append [drive]search-path1[;[drive]search-path2]...\n");
     printf("Usage: append [;]\n\n");

     return 0;
  }


  /* try to find append data file */
  append_file = Find_Append_File();
  #ifdef DEBUG
  if (append_file)
    printf("Found append file.\n");
  else
    printf("Error opeing/creating append file.\n");
  #endif
  
  /* read data from append file (if it exists) */
  memset(&append_struct, 0, sizeof(append_struct) );
  if (append_file)
  {
    status = Read_Append_Data(append_file, &append_struct);
  }


  if (argc < 2)
  {
     /* print out search path */
     #ifdef DEBUG
     printf("Print out search path.\n");
     #endif

     if (append_struct.append_path)
       printf("%s\n", append_struct.append_path);

     done_task = 1;
  }


  if ( Find_Arg("/X", NULL, argc, argv) )
  {
     #ifdef DEBUG
     printf("Found an /x\n");
     #endif

     append_struct.x_flag = 1;
     done_task = 1;
  }

  if ( Find_Arg("/E", NULL, argc, argv) )
  {
     #ifdef DEBUG
     printf("Found an /e\n");
     #endif

     append_struct.environment_flag = 1;
     done_task = 1;
  }

  if (done_task)
    Write_Append_Data(append_file, &append_struct);

  if ( argv[1][0] == ';' )
  {
     /* clear search path */
     #ifdef DEBUG
     printf("Clear that search path.\n");
     #endif

     if (append_struct.append_path)
     {
       free(append_struct.append_path);
       append_struct.append_path = NULL;
     }

     /* write data to append file, if possible */
     if (append_file)
        Write_Append_Data(append_file, &append_struct);

     /* if environment flag is set, save the search path in %APPEND% */
     if (append_struct.environment_flag)
     {
        char cmd_line[1024];
        sprintf(cmd_line, "set append=");
        system(cmd_line);
     }
     done_task = 1;
  }

  /* The new search path must have been given, get and save it. */
  if (!done_task)
  {
     #ifdef DEBUG
     printf("Getting new search path.\n");
     #endif

     if (append_struct.append_path)
       free(append_struct.append_path);

     done_task = 0;     /* just to be proper */
     for (status = 1; status < argc; status++)
       done_task += strlen( argv[status] );

     #ifdef DEBUG
     printf("New search path length: %d\n", done_task);
     #endif

     append_struct.append_path = (char *) calloc(done_task, sizeof(char) );
     if (append_struct.append_path)
     {
        strcpy(append_struct.append_path, argv[1]);
        #ifdef DEBUG
        printf("Path: %s\n", append_struct.append_path);
        #endif

        for (status = 2; status < argc; status++)
        {
          strcat(append_struct.append_path, argv[status]);
          #ifdef DEBUG
          printf("Path: %s\n", append_struct.append_path);
          #endif
        }

        /* now that we have the whole path, write it to the file */
        Write_Append_Data(append_file, &append_struct);

        /* write data to DOS environment */
        if (append_struct.environment_flag)
        {
           char cmd_line[1024];
           sprintf(cmd_line, "set append=%s", append_struct.append_path);
           system(cmd_line);
        }

     }    /* end of calloc worked properly */
  }       /* get new path block */

  /* clean up */
  if (append_file)
  {
     fflush(append_file);
     fclose(append_file);
  }

  if (append_struct.append_path)
    free(append_struct.append_path);

  return 0;
}





/* args.c
Find_Arg()

This function will search the command line for arguments
that match "find_arg". If one is found, the function
returns non-zero and fills the "value" array with the
following argument.
If the argument is not found, then zero is returned
and "value" is undefined.

"find_arg" should have the format "-g" or "-t". A dash followed
by the letter to search for.

This function is case sensitive; "-g" is not the same as "-G".

If the variable "value" is NULL then the argument is not
copied into the value string.

Find_Arg() is written by: Jesse Smith <slicer69@hotmail.com>
http://slicer69.tripod.com/
*/

int Find_Arg(char *find_arg, char *value, int argc, char *argv[])
{
   int count = 1, found = 0;
   
   while ((count < argc) && (!found))
   {
      if ( !strcmp(find_arg, argv[count]) )
         found = 1;
         
      count++;
   }

   if (!found)
      return 0;

   else if ((value) && (argv[count]))
      strcpy(value, argv[count]);

   return 1;
}



/*
This function sets an array of strings to all
upper case. The number of strings in the
array is passed in the second argument.
Jesse <slicer69@hotmail.com>
*/
void Upcase_Strings(char **str, int num_strings)
{
   int string_num, index;
   int string_length;

   for (string_num = 0; string_num < num_strings; string_num++)
   {
      string_length = strlen( str[string_num] );
      for (index = 0; index < string_length; index++)
        str[string_num][index] = toupper( str[string_num][index] );
   }
}




/*
This file reads in data from the append file. It is stored in
the append structure. The file has a simular structure to the
one in memory. However, we must alloc memory for the append
string (path).
This function returns true on success or false on failure.
Jesse <slicer69@hotmail.com>
*/
int Read_Append_Data(FILE *append_file, APPEND_STRUCT *append_struct)
{
   int file_size;
   int status;
   char *path;

   /* get size of the append file */
   fseek(append_file, 0, SEEK_END);
   file_size = (int) ftell(append_file);
   rewind(append_file);

   #ifdef DEBUG
   printf("File size: %d\n", file_size);
   #endif

   /* remove old struct data */
   if (append_struct->append_path)
   {
      free(append_struct->append_path);
      append_struct->append_path = NULL;
   }

   /* create temp path varible, pad with NULL */
   path = (char *) calloc(file_size, sizeof(char) );
   if (!path)
   {
     #ifdef DEBUG
     printf("Memory error, allocing path.\n");
     #endif
     return 0;
   }

   /* read in flags */
   /* the flags all come first in the struct, so we can just read
      them all in at once
   */
   status = fread(append_struct, 2, sizeof(int), append_file);
   if (!status)
   {
     free(path);
     return 0;
   }

   /* read in the append path from the file */
   status = fread(path, file_size, sizeof(char), append_file);
   #ifdef DEBUG
   printf("Path: %s\n", path);
   #endif

   append_struct->append_path = path;
   return 1;
}



/*
This function writes the append_structure to the append_file. If all
goes well, the function returns true, otherwise false is returned.
The flags of the struct are written first, followed by the search
path(s).
Written by: Jesse <slicer69@hotmail.com>
*/
int Write_Append_Data(FILE *append_file, APPEND_STRUCT *append_struct)
{
   int status;

   rewind(append_file);
   status = fwrite(append_struct, 2, sizeof(int), append_file);

   #ifdef DEBUG
   printf("About to write append data to file. %d\n", status);
   #endif

   if (append_struct->append_path)
   {
      #ifdef DEBUG
      printf("Writing: %s\n", append_struct->append_path);
      #endif

      status = fwrite(append_struct->append_path,
                      strlen(append_struct->append_path) + 1,
                      sizeof(char), append_file);
      #ifdef DEBUG
      printf("Return %d from write.\n", status);
      #endif
   }
   else
     fputc('\0', append_file);

   fflush(append_file);

   if (status)
     return 1;
   else
     return 0;
}




/*
This function searches for the append data file. The default
name for this file is stored in the constant APPEND_FILE_NAME.
The function searches directories, in the following order:
C:\
\
.\

If the file cannot be found, the program tries to create the
file in the directories in the above order.

If the file is found (or opened) its handle is returned. If the
file is not found/opened/created, then NULL is returned.
Written by Jesse <slicer69@hotmail.com>
*/
FILE *Find_Append_File()
{
   char path[64];
   FILE *append_file = NULL;

   /* try to find/open file */
   sprintf(path, "c:\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "rb+");
   if (append_file) return append_file;
   sprintf(path, "\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "rb+");
   if (append_file) return append_file;
   sprintf(path, ".\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "rb+");
   if (append_file) return append_file;

   /* if we can't find the file, make a new one */
   sprintf(path, "c:\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "wb+");
   if (append_file) return append_file;
   sprintf(path, "\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "wb+");
   if (append_file) return append_file;
   sprintf(path, ".\\%s", APPEND_FILE_NAME);
   append_file = fopen(path, "wb+");
   if (append_file) return append_file;

   /* couldn't find/create file */
   return NULL;
}
