/*    
   Lowfat32.c - Low level FAT32 api calls (DJGPP version).
   
   Copyright (C) 2000, 2004 Imre Leber

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   If you have any questions, comments, suggestions, or fixes please
   email me at:  imre.leber@worldonline.be
*/

#include <dos.h>
#include <string.h>
#include <dpmi.h>


#define CARY 1

struct DISKIO
{
    unsigned long  diStartSector;     /* Sector number to start.    */
    unsigned short diSectors;         /* Number of sectors.         */
    unsigned long  diBuffer;          /* Segment address of buffer. */
} __attribute__((packed));

struct xdfree
{
    unsigned short size;
    unsigned short version;
    unsigned long  SectorsPerCluster;
    unsigned long  BytesPerSector;
    unsigned long  AvailableClusters;
    unsigned long  TotalClusters;
    unsigned long  AvailablePhysicalSectors;
    unsigned long  TotalPhysicalSectors;
    unsigned long  AvailUnits;
    unsigned long  TotalUnits;
    char           Reserved[8];
} __attribute__((packed));

/******************************************************************
**                     ExtendedAbsReadWrite
*******************************************************************
** Call to extended absolute read/write function in C
**
** This calls needs the FAT32 api, because this may run on a
** non-FAT32 enabled kernel you should first check wether
** the FAT32 api is available.
**
** Also works on FAT12/16.
*******************************************************************/

int ExtendedAbsReadWrite(int drive, int nsects, unsigned long lsect,
			 void* buffer, unsigned area, unsigned BytesPerSector)
{
    int ioseg, bufseg;
    int iosel, bufsel;
    __dpmi_regs regs;
    struct DISKIO io;	
    
    /* Allocate dos memory */
    if ((bufseg = __dpmi_allocate_dos_memory(((nsects * BytesPerSector) + 15) >> 4 , &bufsel)) == -1)
        return -1;  
    
    if ((ioseg = __dpmi_allocate_dos_memory((sizeof(struct DISKIO)+15)>>4, &iosel)) == -1)
    {
        __dpmi_free_dos_memory(bufsel);
	return -1;  
    }
    
    /* Copy buffer to conventional memory */	
    dosmemput(buffer, nsects * BytesPerSector, bufseg*16);
    
    /* Copy diskio struct to conventional memory */
    io.diStartSector = lsect;
    io.diSectors     = nsects;
    io.diBuffer      = bufseg * 16;    
    
    dosmemput(&io, sizeof(struct DISKIO), ioseg*16);
    
    regs.x.ax = 0x7305;
    regs.x.cx = 0xffff;
    regs.x.si = (unsigned short)area;
    regs.h.dl = (unsigned char) drive;
    regs.x.bx = 0;
    regs.x.ds = ioseg;
    
    __dpmi_int(0x21,&regs);    

    dosmemget(bufseg* 16, nsects * BytesPerSector, buffer);
    
    __dpmi_free_dos_memory(bufsel);
    __dpmi_free_dos_memory(iosel);
    
    return (regs.x.flags & CARY) ? -1 : 0;
}

/******************************************************************
**                     GetFAT32SectorSize
*******************************************************************
** Returns the size of a FAT32 sector.
**
** This function can be used to see wether the FAT32 api is available.
*******************************************************************/

unsigned GetFAT32SectorSize(int drivenum)
{
    __dpmi_regs regs;
    int drvseg, drvsel;    
    int freeseg, freesel; 
    char drive[4];
    struct xdfree free;
    unsigned retVal;  
        
    if ((drvseg = __dpmi_allocate_dos_memory(1 ,&drvsel)) == -1)
        return 0;  
    
    if ((freeseg = __dpmi_allocate_dos_memory(((sizeof(struct xdfree))+15)>>4, &freesel)) == -1)
    {
        __dpmi_free_dos_memory(drvsel);
	return 0;  
    }
    
    strcpy(drive, "?:\\");
    drive[0] = (char) drivenum + 'A';
    dosmemput(drive, 4, drvseg*16);
    
    free.version = 0;
    dosmemput(&free, sizeof(struct xdfree), freeseg*16);	
    
    regs.x.ax = 0x7303;    
    
    regs.x.ds = drvseg;
    regs.x.dx = 0;
    
    regs.x.es = freeseg;
    regs.x.di = 0;
    
    regs.x.cx = sizeof(struct xdfree);
	
    __dpmi_int(0x21,&regs);    
    
    if (regs.x.flags & 1)
	retVal = 0;
    else if (regs.h.al == 0)	/* Old DOS */
	retVal = 0;
    else
    {
	dosmemget(freeseg*16, sizeof(struct xdfree), &free);
	retVal = (unsigned) free.BytesPerSector;
    }
    
    __dpmi_free_dos_memory(drvsel);
    __dpmi_free_dos_memory(freesel);

    return retVal;
}

