/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h> /* for mode definitions */

#include "inc\msdos_fs.h"

#define TRUE  1
#define FALSE 0

/* Defines to indicate which part of the disk is being written to,
   match the value expected in SI by FAT32 absolute read/write.    */
#define WR_UNKNOWN   0x0001
#define WR_FAT       0x2001
#define WR_DIRECT    0x4001
#define WR_DATA      0x6001

#define SECTOR unsigned long

#define HD32MB       33554432L       /* Size of 32Mb hard disk. */

static unsigned BytesPerSector;		/* TODO: fill in */

static int IsImageFile;			/* Wether we are working on image files */
static int handle;			/* drive letter - 'A' or file handle */

/* Function to read from drive or image file. */
static int (*ReadFunc)  (int handle, int nsects, SECTOR lsect, void* buffer);
   
/* Function to write to drive or image file. */
static int (*WriteFunc) (int handle, int nsects, SECTOR lsect, void* buffer);

/* Emulated file pointer */
static loff_t VolumePointer;


static int ReadFromRealDisk12(int handle, int nsects, SECTOR lsect,
                              void* buffer);
static int WriteToRealDisk12(int handle, int nsects, SECTOR lsect,
                             void* buffer);
static int ReadFromRealDisk16(int handle, int nsects, SECTOR lsect,
                              void* buffer);	
static int WriteToRealDisk16(int handle, int nsects, SECTOR lsect,
                             void* buffer);	
static int ReadFromRealDisk32(int handle, int nsects, SECTOR lsect,
                              void* buffer);		
static int WriteToRealDisk32(int handle, int nsects, SECTOR lsect,
                             void* buffer);	
			     
/**************************************************************************************
***			OpenVolume
***************************************************************************************
*** Opens a volume (drive or image file)
***************************************************************************************/

int OpenVolume(char* driveorfile)
{    
    if (HasAllFloppyForm(driveorfile))
    {
       handle = toupper(driveorfile[0]) - 'A';

       BytesPerSector = GetFAT32SectorSize(handle);
	
       if (BytesPerSector)
       {   
          ReadFunc  = ReadFromRealDisk32;
	  WriteFunc = WriteToRealDisk32;
       }
       else
       {
	  return -1;   
       }
       
/*       
       else
       {
	  struct DEVICEPARAMS devicepars;
	  unsigned long disksize;

	  /* Get bytes per sector * /
	  if (disk_getparams(handle, &devicepars) != DISK_OK)
	  {
	      return -1;
	  }

	  BytesPerSector = devicepars.sec_size;
	  
	  disksize = BytesPerSector * devicepars.total_sectors;
	  
	  
	  if (disksize < HD32MB)
	  {
	     ReadFunc  = ReadFromRealDisk12;
	     WriteFunc = WriteToRealDisk12;
	  }
	  else
	  {
	     ReadFunc  = ReadFromRealDisk16;
	     WriteFunc = WriteToRealDisk16;
	  }
       }
*/  

    }
    else
    {
       handle      = open (driveorfile, O_RDWR|O_BINARY);
       IsImageFile = TRUE;

       if (handle == -1)
       {
	   return -1;
       }
    }

    return 2;    /* Emulate 2 for the file descriptor */
}

/**************************************************************************************
***			ReadVolume
***************************************************************************************
*** Reads from a volume, making the translation between byte oriented and sector 
*** oriented logic, returns -1 on error
***************************************************************************************/

int ReadVolume(void * data, size_t size)
{
    size_t origsize = size;    
    
    if (IsImageFile)
    {
	/* Just read from file */
	if (lseek(handle, VolumePointer, SEEK_SET) == -1L)
	{
	   return -1;
	}

	if ((read(handle, data, size)) == -1)
	{
	   return -1;      
	}		
    }
    else
    {
	/* Read from real disk */
	unsigned sector = VolumePointer / BytesPerSector;
	unsigned offset = VolumePointer % BytesPerSector;
	unsigned rest   = BytesPerSector - offset;	

	char* buf = (char*) malloc(BytesPerSector);
	if (!buf) 
	    return -1;
	
	if (rest >= size)
	{
	    if (ReadFunc(handle, 1, sector, buf) == -1)
	    {
	        free(buf);
		return -1;
	    }
	   
	    memcpy(data, buf + offset, size);
	}
	else
	{
	    size_t i;
	    char* cdata = (char*) data;
	    
	    if (rest)
	    {
		if (ReadFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;
		}
	   
		memcpy(data, buf + offset, rest);	    
		
		sector++;
	    }
	    
	    cdata += rest;
	    size  -= rest;
	    
	    for (i = 0; i < size / BytesPerSector; i++)
	    {
		if (ReadFunc(handle, 1, sector, cdata) == -1)
		{
		    free(buf);
		    return -1;	    		
		}
	
		cdata += BytesPerSector;
		sector++;
	    }
	    
	    size %= BytesPerSector;
	    if (size)
	    {
		if (ReadFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;
		}
	   
		memcpy(cdata, buf, size);
	    }	    
	}	
	
	free(buf);
    }
    
    return origsize;
}

/**************************************************************************************
***			WriteVolume
***************************************************************************************
*** Writes to a volume, making the translation between byte oriented and sector 
*** oriented logic.
***
*** Returns -1 on error
***************************************************************************************/

int WriteVolume(void * data, size_t size)
{
    int i;
    char* buf, *cdata;
    size_t origsize = size;
    
    if (IsImageFile)
    {
	/* Just write to file */
	if (lseek(handle, VolumePointer, SEEK_SET) == -1L)
	{
	   return -1;
	}

	if ((write(handle, data, size)) == -1)
	{
	   return -1;      
	}	
    }
    else
    {	
	/* Write to real disk */
	unsigned sector = VolumePointer / BytesPerSector;
	unsigned offset = VolumePointer % BytesPerSector;
	unsigned rest   = BytesPerSector - offset;	

	buf = (char*) malloc(BytesPerSector);
	if (!buf) 
	    return -1;
	
	if (rest >= size)
	{
	    if (ReadFunc(handle, 1, sector, buf) == -1)
	    {
	        free(buf);
		return -1;
	    }
	   
	    memcpy(buf + offset, data, size);
	     
	    if (WriteFunc(handle, 1, sector, buf) == -1)
	    {
	        free(buf);
		return -1;	    
	    }
	}
	else
	{
	    cdata = (char*) data;
	    
	    if (rest)
	    {
		if (ReadFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;
		}
	   
		memcpy(buf + offset, data, rest);
	    
		if (WriteFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;	    
		}
		
		sector++;
	    }
	    
	    cdata += rest;
	    size  -= rest;
	    
	    for (i = 0; i < size / BytesPerSector; i++)
	    {
		if (WriteFunc(handle, 1, sector, cdata) == -1)
		{
		    free(buf);
		    return -1;	    		
		}
		
		cdata += BytesPerSector;
		sector++;
	    }
	    
	    size %= BytesPerSector;
	    if (size)
	    {
		if (ReadFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;
		}
	   
		memcpy(buf, cdata, size);
	    
		if (WriteFunc(handle, 1, sector, buf) == -1)
		{
		    free(buf);
		    return -1;	    
		}
	    }	    
	}	
    }
    
    return origsize;
}

/**************************************************************************************
***			CloseVolume
***************************************************************************************
*** Closes a volume, i.e. if this is an image file, then close the file handle.
***************************************************************************************/

int CloseVolume()
{
    if (IsImageFile) 
	return close(handle);
    
    return 0;
}

/**************************************************************************************
***			SeekVolume
***************************************************************************************
*** Sets the virtual file pointer
***************************************************************************************/

off_t VolumeSeek(off_t offset)
{
    VolumePointer = offset;
    return offset;
}

/**************************************************************************************
***			IsWorkingOnImageFile
***************************************************************************************
*** Returns wether we are dealing with an image file
***************************************************************************************/

int IsWorkingOnImageFile()
{
    return IsImageFile;
}

/**************************************************************************************
***			GetVolumeHandle
***************************************************************************************
*** Returns the volume handle
***************************************************************************************/

int GetVolumeHandle()
{
    return handle;
}


/* Real disk sector modifiers. */
/*
   There are different read/write functions:
   - ReadFromRealDisk12, WriteToRealDisk12 are used for FAT12 and FAT16
     volumes <= 32 MB
   - ReadFromRealDisk16, WriteToRealDisk16 are used for FAT16
     volumes > 32 MB
   - ReadFromRealDisk32, WriteToRealDisk32 are used for systems supporting the FAT32 API,
     (like FreeDOS)

   Notice also that, despite the names absread/abswrite work independent
   of the file system used (i.e. the naming here is a litle unfortunate)
*/		      
/*
static int ReadFromRealDisk12(int handle, int nsects, SECTOR lsect,
                              void* buffer)
{
  if (disk_read(handle, (int) lsect, buffer, nsects, BytesPerSector) != DISK_OK)
  {
     return -1;
  }

  return 0;
}

static int WriteToRealDisk12(int handle, int nsects, SECTOR lsect,
                             void* buffer)
{
  if (disk_write(handle, (int) lsect, buffer, nsects, BytesPerSector) != DISK_OK)
  {
     return -1;
  }

  return 0;
}

static int ReadFromRealDisk16(int handle, int nsects, SECTOR lsect,
                              void* buffer)
{
   if (disk_read_ext(handle, lsect, buffer, nsects, BytesPerSector) != DISK_OK)
   {
     return -1;
   }
   return 0;
}

static int WriteToRealDisk16(int handle, int nsects, SECTOR lsect,
                             void* buffer)
{
   if (disk_write_ext(handle, lsect, buffer, nsects, BytesPerSector) != DISK_OK)
   {
     return -1;
   }
   return 0;
}
*/

static int ReadFromRealDisk32(int handle, int nsects, SECTOR lsect,
                              void* buffer)
{
   /* Don't use DISKLIB here (it uses ioctl for FAT12/16 which according to
      microsoft docs does not work!) */

   return ExtendedAbsReadWrite(handle+1, nsects, lsect, buffer, 0, BytesPerSector);
}

static int WriteToRealDisk32(int handle, int nsects, SECTOR lsect,
                             void* buffer)
{
   unsigned area = WR_UNKNOWN;	/* TODO: determine area? */
    
   return ExtendedAbsReadWrite(handle+1, nsects, lsect, buffer, area, BytesPerSector);
}


